//
//  TDLCSVAtomicStore.m
//  ModernToDoList
//
//  Created by narita on 2017/01/16.
//  Copyright © 2017年 narita. All rights reserved.
//

#import "TDLCSVAtomicStore.h"

NSString * const	TDLCSVAtomicStoreType		= @"TDLCSVAtomicStoreType";

@interface TDLCSVAtomicStore ()

@property (strong) NSMutableDictionary *nodeCacheRef;

@end

@implementation TDLCSVAtomicStore

#pragma mark -
#pragma mark init methods

- (id)initWithPersistentStoreCoordinator:(NSPersistentStoreCoordinator *)coordinator
                       configurationName:(NSString *)configurationName 
                                     URL:(NSURL *)anUrl
                                 options:(NSDictionary *)options 
{
    if(nil == anUrl || [anUrl isFileURL] == NO ) 
    {
        self = nil;
    }
    else
    {
        self = [super initWithPersistentStoreCoordinator:coordinator 
                                       configurationName:configurationName 
                                                     URL:anUrl 
                                                 options:options];
        
        if( self != nil )
        {
            NSFileManager* theFileManager = [NSFileManager defaultManager];
            NSString*      theFilePath    = [anUrl path];
            
            BOOL           isDirectory;
            BOOL           isFileExist;
            
            isFileExist = [theFileManager fileExistsAtPath:theFilePath isDirectory:&isDirectory];
            
            if( !isFileExist )
            {
                // ファイルが存在しないが作れる
                if( [theFileManager createFileAtPath:theFilePath contents:nil attributes:nil] == NO )
                {
                    self = nil;
                }
                else
                {
                    NSDictionary *metadata = [TDLCSVAtomicStore metadataForPersistentStoreWithURL:[self URL] 
                                                                                         error:nil];
                    [self setMetadata:metadata];
                    
                    self.nodeCacheRef = [NSMutableDictionary dictionary];
                }
                
            } 
            else if (isDirectory)
            {
                self = nil;// ディレクトリはサポート外
            }
            else
            {
                // ファイルが存在するのでOK
                NSDictionary *metadata = [TDLCSVAtomicStore metadataForPersistentStoreWithURL:[self URL] 
                                                                                     error:nil];
                [self setMetadata:metadata];
                
                self.nodeCacheRef = [NSMutableDictionary dictionary];
            }
        }
    }
        
    return self;
}

#pragma mark -
#pragma mark Metadata

+ (NSDictionary *)metadataForPersistentStoreWithURL:(NSURL *) url
                                              error:(NSError *__autoreleasing *) error 
{
    NSString* theFilePath = [url relativePath];
    NSMutableDictionary* theMetadata = [NSMutableDictionary dictionary];

    // ファイルが存在しないならばdefaultの値を生成する
    if(![[NSFileManager defaultManager] fileExistsAtPath:theFilePath]) 
    {        
        [theMetadata setValue:[NSUUID UUID].UUIDString 
                       forKey:NSStoreUUIDKey];
        
        // ファイルに書き込む
        [TDLCSVAtomicStore setMetadata:theMetadata
             forPersistentStoreWithURL:url 
                                 error:nil];
    }
    
    // ファイルから1行目を読み込んでその値をuuidとする
    NSString* theContents = [NSString stringWithContentsOfURL:url
                                                     encoding:NSUTF8StringEncoding
                                                        error:nil];
    // 行"\n"で分解
    NSArray* theLines = [theContents componentsSeparatedByString:@"\n"];
    
    if( theLines.count > 0 )
    {
        NSString* theValue = [NSUUID UUID].UUIDString;// デフォルトち
        
        if( [theLines[0] length] > 1)
        {
            theValue = [ theLines[0] substringFromIndex:1];
        }
        
        // 最初の行から最初の一文字の#を外した文字列がUUID
        [theMetadata setValue:theValue
                       forKey:NSStoreUUIDKey]; 
    }

    // 定数の書き込み
    [theMetadata setValue:TDLCSVAtomicStoreType
                   forKey:NSStoreTypeKey]; 
    
    return theMetadata;
}

+ (BOOL)        setMetadata:(NSDictionary *) metadata 
  forPersistentStoreWithURL:(NSURL*) url 
                      error:(NSError *__autoreleasing *) error 
{
    NSString* theFilePath = [url relativePath];

    // ファイルが存在しないならばdefaultの値を生成する
    if(![[NSFileManager defaultManager] fileExistsAtPath:theFilePath]) 
    {    
        NSString* theDummyString = @"#DUMMY\n";
        
        [theDummyString writeToFile:theFilePath
                         atomically:YES
                           encoding:NSUTF8StringEncoding
                              error:error];
    }

    // ファイルを読み込んで、
    // ファイルから1行目を読み込んでその値をuuidとする
    NSString* theContents = [NSString stringWithContentsOfURL:url
                                                     encoding:NSUTF8StringEncoding
                                                        error:nil];
    // 行"\n"で分解
    NSArray* theLines = [theContents componentsSeparatedByString:@"\n"];
    
    // #がある最初の1行目を削除して、新たに1行目位行こうを追加して
    
    // 書き出すファイルの内容を貯めておく場所
    NSMutableString* theOutputString = [@"" mutableCopy];
    
    // メタデータを書き出す
    [theOutputString appendFormat:@"#%@\n", [metadata objectForKey:NSStoreUUIDKey]];

    for(NSString* i in [theLines subarrayWithRange:NSMakeRange(1, theLines.count-1)])
    {
        [theOutputString appendFormat:@"%@\n", i];
    }

    [theOutputString writeToFile:theFilePath
                      atomically:YES
                        encoding:NSUTF8StringEncoding
                           error:nil];

    return YES;
}

- (NSString *)type
{
    return [self.metadata objectForKey:NSStoreTypeKey];
}

- (NSString *)identifier 
{
    return [self.metadata objectForKey:NSStoreUUIDKey];
}


#pragma mark -
#pragma mark Loading

- (BOOL)load:(NSError *__autoreleasing *)__unused error 
{
    // ファイルから読み込み    
    NSString* theContents = [NSString stringWithContentsOfURL:[self URL]
                                                     encoding:NSUTF8StringEncoding
                                                        error:error];

    if( *error != nil )
    {
        return NO;
    }
    
    NSMutableSet* theCacheNodes = [NSMutableSet set];

    
    NSEntityDescription* theEntity =[[[[self persistentStoreCoordinator] managedObjectModel]
                                      entitiesByName] objectForKey:@"ToDoItem"];// <-- ハードコーディングする

    // 行"\n"で分解
    NSArray* theLines = [theContents componentsSeparatedByString:@"\n"];
    
    // 最初の1行目だけ特別扱いする
    if( theLines.count > 0 )
    {
        NSString* theValue = [NSUUID UUID].UUIDString;// デフォルトち
        
        if( [theLines[0] length] > 1)
        {
            theValue = [ theLines[0] substringFromIndex:1];
        }
        
        // 最初の行から最初の一文字の#を外した文字列がUUID
        [self.metadata setValue:theValue
                       forKey:NSStoreUUIDKey]; 
    }    
    
    for(NSString* theLine in [theLines subarrayWithRange:NSMakeRange(1, theLines.count-1)])
    {
        NSString* theItem;
        // ","で分解
        NSArray* theItems = [theLine componentsSeparatedByString:@","];

        if( theItems.count == 5 )
        {
            // 前後のスペースを削除
            theItem = [theItems[4] stringByTrimmingCharactersInSet:[NSCharacterSet whitespaceCharacterSet]];
            
            // ノードの作成
            NSManagedObjectID *objectID = [self objectIDForEntity:theEntity
                                                  referenceObject:theItem];
            NSAtomicStoreCacheNode *theNewNode = [self nodeForReferenceObject:theItem  
                                                                    andObjectID:objectID];            
            
            // 前から、done, memo, sortOrder, createDateとする
            theItem = [theItems[0] stringByTrimmingCharactersInSet:[NSCharacterSet whitespaceCharacterSet]];
            [theNewNode setValue:[NSNumber numberWithBool:[theItem boolValue]]
                          forKey:@"done"];
            
            theItem = [theItems[1] stringByTrimmingCharactersInSet:[NSCharacterSet whitespaceCharacterSet]];
            [theNewNode setValue:[theItem mutableCopy]
                          forKey:@"memo"];
            
            theItem = [theItems[2] stringByTrimmingCharactersInSet:[NSCharacterSet whitespaceCharacterSet]];
            [theNewNode setValue:[NSNumber numberWithInt:[theItem intValue]]
                          forKey:@"sortOrder"];
            
            theItem = [theItems[3] stringByTrimmingCharactersInSet:[NSCharacterSet whitespaceCharacterSet]];
            [theNewNode setValue:[NSDate dateWithString:theItem]
                          forKey:@"createdDate"];
            
            [theCacheNodes addObject:theNewNode];
        }
    }
    
    [self addCacheNodes: theCacheNodes];
    
    return YES;
}

#pragma mark -
#pragma mark Saving

- (BOOL)save:(NSError *__autoreleasing *) error 
{
    // 書き出すファイルの内容を貯めておく場所
    NSMutableString* theOutputString = [@"" mutableCopy];
    
    // メタデータを書き出す
    [theOutputString appendFormat:@"#%@\n", [self.metadata objectForKey:NSStoreUUIDKey]];
    
    // ノード全部を書き出す。
    NSSet* theNodes = [self cacheNodes];
    
    // todo:sortOrderでソートしておくと人間が見るときに便利かも
    
    // TDLアプリ専用なので、ハードコーディングでOK
    for( NSAtomicStoreCacheNode* theNode in theNodes )
    {        
        [theOutputString appendFormat:@"%@, %@, %@, %@, %@\n", 
         [theNode valueForKey:@"done"],
         [theNode valueForKey:@"memo"],
         [theNode valueForKey:@"sortOrder"],
         [theNode valueForKey:@"createdDate"],
         [self referenceObjectForObjectID:[theNode objectID]]];
    }

    NSString* theFilePath = [self.URL relativePath];
    [theOutputString writeToFile:theFilePath
                      atomically:YES
                        encoding:NSUTF8StringEncoding
                           error:error];
    
    return YES;
}

#pragma mark -
#pragma mark Chache node management 

- (NSAtomicStoreCacheNode *)newCacheNodeForManagedObject:(NSManagedObject*) inManagedObject
{
    NSManagedObjectID* theObjectID = [inManagedObject objectID];
    
    // リファレンス文字列の作成
    id theReferenceID = [self referenceObjectForObjectID:theObjectID];
    
    // キャッシュノードの作成
    NSAtomicStoreCacheNode *theCacheNode = [self nodeForReferenceObject:theReferenceID  
                                                            andObjectID:theObjectID];
    
    [self updateCacheNode:theCacheNode
        fromManagedObject:inManagedObject];
    
    return theCacheNode;
}

// KVCでmanagedObjectからキャッシュへ値を反映させる
// ここでは、このアプリでしか使わないので、ハードコーディング
-(void)updateCacheNode:(NSAtomicStoreCacheNode *)node
     fromManagedObject:(NSManagedObject *)managedObject 
{
    // TDLアプリ専用なので、ハードコーディングでOK
    [node setValue:[managedObject valueForKey:@"done"] 
            forKey:@"done"];
    
    [node setValue:[managedObject valueForKey:@"memo"] 
            forKey:@"memo"];
    
    [node setValue:[managedObject valueForKey:@"sortOrder"] 
            forKey:@"sortOrder"];
    
    [node setValue:[managedObject valueForKey:@"createdDate"] 
            forKey:@"createdDate"];
    
    return;    
}

- (id)newReferenceObjectForManagedObject:(NSManagedObject *) managedObject 
{    
    return [NSUUID UUID].UUIDString;
}

- (NSAtomicStoreCacheNode *)nodeForReferenceObject:(id)inReference  
                                       andObjectID:(NSManagedObjectID *)inObjectID
{
    NSAtomicStoreCacheNode* theNode = [self.nodeCacheRef objectForKey:inReference];
    
    if( theNode == nil ) 
    {
        theNode = [[NSAtomicStoreCacheNode alloc] initWithObjectID:inObjectID];
        
        [self.nodeCacheRef setObject:theNode
                              forKey:inReference];
    }
    
    return theNode;
}

#pragma mark -
#pragma mark Remove objects

- (void)willRemoveCacheNodes:(NSSet *)cacheNodes 
{
    for(NSAtomicStoreCacheNode *node in cacheNodes) 
    {
        id theReferenceID = [self referenceObjectForObjectID:node.objectID];
        [self.nodeCacheRef removeObjectForKey:theReferenceID];
    }
    
    [super willRemoveCacheNodes:cacheNodes];
}

- (void)willRemoveFromPersistentStoreCoordinator:(NSPersistentStoreCoordinator *)coordinator
{
    // 全てのキャッシュを削除する
    [self willRemoveCacheNodes:[self cacheNodes]];
    self.nodeCacheRef = nil;
    
    [super willRemoveFromPersistentStoreCoordinator:coordinator];
}

#pragma mark - utility metods

@end
