//
//  TDLViewController.m
//  ModernToDoList
//
//  Created by narita on 2016/09/29.
//  Copyright © 2016年 narita. All rights reserved.
//

#import "TDLViewController.h"
#import "ToDoItem+CoreDataClass.h"
#import "ToDoItem_coding.h"
#import "NSIndexSet_Category.h"

@interface TDLViewController ()

@property (strong, readwrite) NSFetchedResultsController* fetchResultController;
@property (readonly) NSManagedObjectContext* managedObjectContext;

- (NSArray<__kindof ToDoItem *>*) sortedTodoItemsWithPredicate:( nullable NSPredicate*)inPredicate;

@end

@implementation TDLViewController

#pragma mark -
#pragma mark class methods

#pragma mark -
#pragma mark init methods

#pragma mark -
#pragma mark dealloc

#pragma mark -
#pragma mark NSCopying, hash, isEqual:

#pragma mark -
#pragma mark NSCoding methods

#pragma mark -
#pragma mark state restore methods

+ (NSArray *)restorableStateKeyPaths
{
    return [[super restorableStateKeyPaths] arrayByAddingObjectsFromArray:@[ @"selectedIndex"]];
}

- (void)restoreStateWithCoder:(NSCoder *)coder
{
    [super restoreStateWithCoder:coder];
    
    [self.tableView selectRowIndexes:self.selectedIndex
                byExtendingSelection:NO];
}

#pragma mark -
#pragma mark awakeFromNib-like methods

- (void)viewDidLoad 
{
    [super viewDidLoad];
    
    // Drag & Drop の設定
    // Drag & drop はアニメーション表示させる
    self.tableView.draggingDestinationFeedbackStyle = NSTableViewDraggingDestinationFeedbackStyleRegular;
    
    [self.tableView registerForDraggedTypes:@[[ToDoItem pasteBoardRowsType]]];
    
    // ゴミ箱に入れる機能を追加するために、NSDragOperationDeleteを加える
    [self.tableView setDraggingSourceOperationMask:( NSDragOperationCopy | 
                                                    NSDragOperationDelete )
                                          forLocal:NO];
    
    [self.tableView setDraggingSourceOperationMask:(NSDragOperationMove | 
                                                    NSDragOperationCopy )
                                          forLocal:YES];
}

- (void)viewWillAppear
{
    [super viewWillAppear];
    
    // ここで設定しないと、self.view.windowがnilの時になり取得できない
    self.representedObject = self.managedObjectContext;
    
    // 表示と編集のための、NSFetchRequestControllerの設定
    if( self.fetchResultController == nil )
    {
        NSFetchRequest<ToDoItem *>* theFetchRequest = [ToDoItem fetchRequest];
        
        theFetchRequest.sortDescriptors = @[[NSSortDescriptor sortDescriptorWithKey:@"sortOrder" ascending:YES]];
        
        self.fetchResultController = 
        [[NSFetchedResultsController alloc] initWithFetchRequest:theFetchRequest
                                            managedObjectContext:self.managedObjectContext
                                              sectionNameKeyPath:nil 
                                                       cacheName:nil];
        
        self.fetchResultController.delegate = self;
    }
        
    [self.fetchResultController performFetch:nil];
    [self.tableView reloadData];
}

- (void) renumberingSortOrder
{
    NSArray* theArray = [self sortedTodoItemsWithPredicate:nil];

    for( ToDoItem* i in theArray )
    {
        i.sortOrder = (int32_t)[theArray indexOfObject:i];
    }
}

#pragma mark -
#pragma mark action methods

- (IBAction) newItem:(nullable id)sender
{
    // 選択位置の次の行に挿入する。選択位置がない場合は末尾に追加する。    
    NSInteger theSelected = self.tableView.selectedRow;
    NSInteger theCount = self.fetchResultController.fetchedObjects.count;

    if( theSelected > -1 && theCount != 0 )
    {                
        NSArray* theArray = [self sortedTodoItemsWithPredicate:
                             [NSPredicate predicateWithFormat:@"sortOrder > %i", theSelected]];
        
        // 選択位置+1から末尾まで+1ずらす
        for( ToDoItem* i in theArray )
        {
            i.sortOrder = i.sortOrder + 1;
        }
        
        // 選択位置+1の位置に追加
        ToDoItem* theItem = [[ToDoItem alloc] initWithContext:self.managedObjectContext];
        
        theItem.memo = @"insertCurrent";
        theItem.sortOrder = (int32_t)(theSelected + 1);
        
        
        // 追加したitemを選択
        [[NSOperationQueue mainQueue] addOperationWithBlock:^{
            [self.tableView selectRowIndexes:[NSIndexSet indexSetWithIndex:theSelected + 1]
                        byExtendingSelection:NO];
        }];    

    }
    else
    {
        // 末尾に追加        
        theSelected = theCount;

        ToDoItem* theItem = [[ToDoItem alloc] initWithContext:self.managedObjectContext];

        theItem.memo = @"addLast";
        theItem.sortOrder = (int32_t)(theSelected);
        
        [self renumberingSortOrder];
        
        // 追加したitemを選択
        [[NSOperationQueue mainQueue] addOperationWithBlock:^{
            [self.tableView selectRowIndexes:[NSIndexSet indexSetWithIndex:(theSelected)]
                        byExtendingSelection:NO];
        }];    
    }
}

- (IBAction) delete:(nullable id)sender
{    
    
    // 選択されていなければ、削除しない
    if( [self canDelete:sender] )
    {        
        NSArray*     theSelectedIndexArray = self.tableView.selectedRowIndexes.array;
        NSPredicate* thePredicate = [NSPredicate predicateWithFormat:@"sortOrder IN %@", theSelectedIndexArray];        
        NSArray*     theArray = [self sortedTodoItemsWithPredicate:thePredicate];

        for(ToDoItem* i in theArray)
        {
            [self.managedObjectContext deleteObject:i];            
        }
        
        [self renumberingSortOrder];
        
//        NSPersistentStoreCoordinator
//        NSAsynchronousFetchRequest* theResuest;
//        NSProgress* theProgress;
        
//        NSFetchRequest *request = [ToDoItem fetchRequest];
//        NSArray*     theSelectedIndexArray = self.tableView.selectedRowIndexes.array;
//        NSPredicate* thePredicate = [NSPredicate predicateWithFormat:@"sortOrder IN %@", theSelectedIndexArray];
//        request.predicate = thePredicate;
//        NSBatchDeleteRequest *delete = [[NSBatchDeleteRequest alloc] initWithFetchRequest:request];
//        
//        [self.managedObjectContext.persistentStoreCoordinator executeRequest:delete 
//                                                                 withContext:self.managedObjectContext 
//                                                                       error:nil];
//        [self renumberingSortOrder];

    }
}

// sortOrderの変更はNSBatchUpdateRequestが使えそう。ここで、バッチ処理にNSExpressionが使用できるので、そこで"sortOrder = self.sortOrder+n"が可能と思われる。


- (BOOL) canDelete:(id)sender
{    
    return self.tableView.numberOfSelectedRows != 0;
}

- (IBAction) dump:(id)sender
{        
//    NSLog(@"objects => %@", self.fetchResultController.fetchedObjects);    
//    [self renumberingSortOrder];
    [self.fetchResultController performFetch:nil];
    [self.tableView reloadData];
}

- (IBAction)copy:(id)sender
{
    // 選択されていなければ、コピーしない
    if( [self canCopy:sender] )
    {
        NSArray*     theSelectedIndexArray = self.tableView.selectedRowIndexes.array;
        NSPredicate* thePredicate = [NSPredicate predicateWithFormat:@"sortOrder IN %@", theSelectedIndexArray];        
        NSArray*     theArray = [self sortedTodoItemsWithPredicate:thePredicate];
        
        NSPasteboard* thePb = [NSPasteboard generalPasteboard];
        
        [thePb clearContents];
        [thePb writeObjects:theArray];
    }
}

- (BOOL) canCopy:(id)sender
{    
    return self.tableView.numberOfSelectedRows != 0;
}

- (IBAction)cut:(id)sender
{
    if( [self canCut:sender] )
    {
        [self copy:self];
        [self delete:self];
    }
}

- (BOOL) canCut:(id)sender
{    
    return self.tableView.numberOfSelectedRows != 0;
}

- (IBAction)paste:(id)sender
{
    // 選択位置の次の行に挿入する。選択位置がない場合は末尾に追加する。    
    NSInteger theSelected = self.tableView.selectedRow;
    
    if( theSelected > -1 )
    {        
        NSArray* theArray = [self sortedTodoItemsWithPredicate:
                             [NSPredicate predicateWithFormat:@"sortOrder > %i", theSelected]];
        
        // 選択位置+1から末尾まで+1ずらす
        for( ToDoItem* i in theArray )
        {
            i.sortOrder = i.sortOrder + 1;
        }
        
        // 選択位置+1の位置に追加
        ToDoItem* theItem = [ToDoItem toDoItemFromNSPasteboard:[NSPasteboard generalPasteboard]
                                          managedObjectContext:self.managedObjectContext];
        theItem.sortOrder = (int32_t)(theSelected + 1);
        
        // 追加したitemを選択
        [[NSOperationQueue mainQueue] addOperationWithBlock:^{
            [self.tableView selectRowIndexes:[NSIndexSet indexSetWithIndex:theSelected + 1]
                        byExtendingSelection:NO];
        }];    
        
    }
    else
    {
        // 末尾に追加        
        theSelected = self.fetchResultController.fetchedObjects.count;
        
        ToDoItem* theItem = [ToDoItem toDoItemFromNSPasteboard:[NSPasteboard generalPasteboard]
                                          managedObjectContext:self.managedObjectContext];        
        theItem.sortOrder = (int32_t)(theSelected);
        
        [self renumberingSortOrder];
        
        // 追加したitemを選択
        [[NSOperationQueue mainQueue] addOperationWithBlock:^{
            [self.tableView selectRowIndexes:[NSIndexSet indexSetWithIndex:(theSelected)]
                        byExtendingSelection:NO];
        }];    
    }
}

- (BOOL) canPaste:(id)sender
{    
    NSString* theType = [[NSPasteboard generalPasteboard] 
                         availableTypeFromArray:@[[ToDoItem pasteBoardRowsType]]];
    
    return theType != nil;
}

- (BOOL)validateUserInterfaceItem:(id<NSValidatedUserInterfaceItem>)anItem
{
    BOOL theResult = NO;
    
    SEL theSelector = [anItem action];
    
    if( theSelector == @selector(copy:) )
    {
        theResult = [self canCopy:anItem];
    }
    else if( theSelector == @selector(cut:) )
    {
        theResult = [self canCut:anItem];
    }
    else if( theSelector == @selector(paste:) )
    {
        theResult = [self canPaste:anItem];
    }
    else if( theSelector == @selector(delete:))
    {
        theResult = [self canDelete:anItem];
    }
    
    return theResult;
}


#pragma mark -
#pragma mark delegate/datasource methods

#pragma mark -
#pragma mark NSTableView delegate

// 選択位置の変更通知を取得して、状態の復帰用の変数を更新しておく。
- (void)tableViewSelectionDidChange:(NSNotification *)notification
{
    self.selectedIndex = self.tableView.selectedRowIndexes;
}

#pragma mark -
#pragma mark NSTableView datasource

// 表示するデータの個数
- (NSInteger)numberOfRowsInTableView:(NSTableView *)tableView
{
    return self.fetchResultController.fetchedObjects.count;
}

// 表示する個々のデータを行ごとに渡す。
// この場所で、クラスToDoItemが、カスタムクラスTDLToDoItemCellViewのobjectValueへ代入される。
- (nullable id) tableView:(NSTableView *)tableView 
objectValueForTableColumn:(nullable NSTableColumn *)tableColumn 
                      row:(NSInteger)row
{
    ToDoItem* theItem = [self.fetchResultController.fetchedObjects objectAtIndex:row];
    
    return theItem;    
}

#pragma mark -
#pragma mark NSTableView Drag & Drop methods

// Drag元のデータをペーストボードにコピー
- (nullable id <NSPasteboardWriting>)tableView:(NSTableView *)inTableView pasteboardWriterForRow:(NSInteger)inRow
{       
    NSArray* theArray = [self sortedTodoItemsWithPredicate:
                         [NSPredicate predicateWithFormat:@"sortOrder == %i", inRow]];
        
    return [theArray firstObject];
}

// ゴミ箱へDropして削除する機能の実装
- (void)tableView:(NSTableView *)tableView 
  draggingSession:(NSDraggingSession *)session 
     endedAtPoint:(NSPoint)screenPoint 
        operation:(NSDragOperation)operation 
{    
    if( operation == NSDragOperationDelete )
    {        
        [session enumerateDraggingItemsWithOptions:NSDraggingItemEnumerationConcurrent
                                           forView:tableView
                                           classes:[NSArray arrayWithObject:[NSPasteboardItem class]]
                                     searchOptions:[NSDictionary dictionary]
                                        usingBlock:^(NSDraggingItem *draggingItem, NSInteger idx, BOOL *stop)
         {
             NSPasteboardItem* thePasteboradItem = draggingItem.item;
             
             if( ![[thePasteboradItem types] containsObject:[ToDoItem pasteBoardRowsType]] )
             {
                 *stop = YES;
             }
             else
             {
                 NSDictionary* thePasteObject = [draggingItem.item propertyListForType:[ToDoItem pasteBoardRowsType]]; 
                 
                 NSURL* theURL = [NSURL URLWithString:[thePasteObject objectForKey:@"objectURL"]];
                                  
                 NSManagedObjectID* theManagedObjectID = 
                 [self.managedObjectContext.persistentStoreCoordinator managedObjectIDForURIRepresentation:theURL];
                                                  
                 [self.managedObjectContext deleteObject:
                  [self.managedObjectContext objectWithID:theManagedObjectID]];            
             }             
         }];    
        
        [self renumberingSortOrder];
    }
}

// Drop可能か否かのチェック
- (NSDragOperation)tableView:(NSTableView*)inTableView
                validateDrop:(id<NSDraggingInfo> )inInfo
                 proposedRow:(NSInteger)inRow
       proposedDropOperation:(NSTableViewDropOperation)inOp
{
    NSDragOperation theResult = NSDragOperationNone;
    
    switch (inOp)
    {
        case NSTableViewDropAbove:// 行と行の間
            if( inTableView == [inInfo draggingSource] ) 
            {
                if([[NSApp currentEvent] modifierFlags] & NSEventModifierFlagOption)
                {
                    theResult = NSDragOperationCopy;
                }
                else
                {
                    // 同じコンテキストならば移動
                    theResult = NSDragOperationMove;        
                }
            }
            else
            {
                // それ以外はコピー
                theResult = NSDragOperationCopy;        
            }
            break;
            
        case NSTableViewDropOn: // 行の上
            theResult = NSDragOperationNone;        
            break;
            
        default:
            theResult = NSDragOperationNone;
            break;
    }
        
    return theResult & [inInfo draggingSourceOperationMask];
}

// Dropの処理
- (BOOL)tableView:(NSTableView*)tv
       acceptDrop:(id<NSDraggingInfo>)info
              row:(NSInteger)row
    dropOperation:(NSTableViewDropOperation)op
{
    NSDragOperation theDragOperation = [self tableView:tv
                                          validateDrop:info
                                           proposedRow:row
                                 proposedDropOperation:op];
        
    if(theDragOperation == NSDragOperationCopy)
    {
        // 1. コピーする個数を確認
        NSInteger theCopyCount   = info.numberOfValidItemsForDrop;
                
        NSArray* theArray = [self sortedTodoItemsWithPredicate:
                             [NSPredicate predicateWithFormat:@"sortOrder > %i", row]];

        
        // 2. インサートする場所以降をコピー個数分だけ増やしておく
        for( ToDoItem* i in theArray )
        {
            i.sortOrder = i.sortOrder + (int32_t)theCopyCount;
        }
                
        // 3. インサートしながらrow+カウントでsortOrderを書き込む        
        [info enumerateDraggingItemsWithOptions:NSDraggingItemEnumerationConcurrent
                                        forView:tv
                                        classes:[NSArray arrayWithObject:[NSPasteboardItem class]]
                                  searchOptions:[NSDictionary dictionary]
                                     usingBlock:^(NSDraggingItem *draggingItem, NSInteger idx, BOOL *stop)
         {
             NSPasteboardItem* thePasteboradItem = draggingItem.item;
             
             if( ![[thePasteboradItem types] containsObject:[ToDoItem pasteBoardRowsType]] )
             {
                 *stop = YES;
             }
             else
             {
                 ToDoItem* theNewObject;
                 theNewObject = [ToDoItem toDoItemFromNSPasteboard:draggingItem.item
                                              managedObjectContext:self.managedObjectContext];
                 theNewObject.sortOrder = (int32_t)(row+idx+1);
             }             
         }];    
                
        return  YES;
    }
    else if(theDragOperation == NSDragOperationMove)
    {
        // コピー＆ペースト&デリートではなくて実体のsortOrderを変更する事。
        // そうしないとObjectIDが変わるのでインスペクターパネルやKVOの関連が切れてしまう。
        
        
        return  NO;        
    }
    else
    {
        return  NO;
    }
    
    return  NO;
}

#pragma mark -
#pragma mark NSFetchedResultsController methods

// 変更通知が来たら、viewを更新させる
- (void)controller:(NSFetchedResultsController *)controller
   didChangeObject:(id)anObject 
       atIndexPath:(nullable NSIndexPath *)indexPath 
     forChangeType:(NSFetchedResultsChangeType)type 
      newIndexPath:(nullable NSIndexPath *)newIndexPath
{
    NSIndexSet* theOldIndexSet = [NSIndexSet indexSetWithIndex:[indexPath indexAtPosition:1]];
    NSIndexSet* theNewIndexSet = [NSIndexSet indexSetWithIndex:[newIndexPath indexAtPosition:1]];
    
    switch(type)
    {
        case NSFetchedResultsChangeInsert:
        {
            [self.tableView insertRowsAtIndexes:theNewIndexSet
                                  withAnimation:NSTableViewAnimationEffectGap];
        }
            break;
            
        case NSFetchedResultsChangeDelete:
        {
            [self.tableView removeRowsAtIndexes:theOldIndexSet
                                  withAnimation:NSTableViewAnimationEffectGap];
        }
            break;
            
        case NSFetchedResultsChangeMove:
        {
            [self.tableView removeRowsAtIndexes:theOldIndexSet
                                  withAnimation:NSTableViewAnimationEffectGap];
            [self.tableView insertRowsAtIndexes:theNewIndexSet
                                  withAnimation:NSTableViewAnimationEffectGap];
        }   
            break;

        case NSFetchedResultsChangeUpdate:
        {
            [self.tableView reloadDataForRowIndexes:theOldIndexSet
                                      columnIndexes:[NSIndexSet indexSetWithIndex:0]];
        }
            break;
            
        default:
            // 想定外の更新情報が来た場合は、強制的に再読み込みする。
            [self.fetchResultController performFetch:nil];
            [self.tableView reloadData];
            break;
    }
}

// 以下の2つのメソッド(beginUpdates,endUpdates)はアニメーションをさせるのに必要
- (void)controllerWillChangeContent:(NSFetchedResultsController *)controller
{
    [self.tableView beginUpdates];
}

- (void)controllerDidChangeContent:(NSFetchedResultsController *)controller
{
    [self.tableView endUpdates];
}

#pragma mark -
#pragma mark accessor methods (in pairs)

- (void) setSelectedIndex:(NSIndexSet *)selectedIndex
{
    _selectedIndex = [selectedIndex copy];
    [self invalidateRestorableState];
}

- (void)setRepresentedObject:(id)representedObject 
{
    [super setRepresentedObject:representedObject];    
}

- (NSManagedObjectContext*) managedObjectContext
{
    NSPersistentDocument* theDoc = self.view.window.windowController.document; 
    NSManagedObjectContext* theContext = theDoc.managedObjectContext;
    
    return theContext;
}

#pragma mark -
#pragma mark Utility methods

- (NSArray<__kindof ToDoItem *>*) sortedTodoItemsWithPredicate:( nullable NSPredicate*)inPredicate
{
    NSFetchRequest* theRquest = [ToDoItem fetchRequest];
    NSSortDescriptor* theSortDescriptor = [[NSSortDescriptor alloc] initWithKey:@"sortOrder" 
                                                                      ascending:YES];
    theRquest.sortDescriptors = @[theSortDescriptor];
    theRquest.predicate = inPredicate;
    
    NSArray* theArray = [self.managedObjectContext executeFetchRequest:theRquest 
                                                                 error:nil];

    return theArray;
}

@end
